/*
Dies ist die Firmware in Hex1.
Hex1 kennt 4 Verhaltensweisen, die durch eine FSM gesteuert werden:
NORMAL
  Hex1 geht solange gerade aus, bis er auf ein Hinternis stt.
AVOIDRIGHT
  Hex1 geht rckwrts und dreht dann nach links um an dem Hinternis vorbei
  zu laufen.
AVOIDLEFT
  Hex1 geht rckwrts und dreht dann nach rechts um an dem Hinternis vorbei
  zu laufen.
PANIC
  Wenn die vorderen Fhler sehr hufig angeschlagen werden, dann dreht sich
  Hex1 um 180 und rennt weg.

Um die Bewegungen zu koordinieren wird eine zweite FSM verwendet, die folgende
Zustnde kennt:
STOP
  Das ist die Stellung mit dem geringsten Energiebedarf.
FORW
  Vorwrts gehen.
BACKW
  Rckwrts gehen.
ROTR
  Nach Rechts rotieren.
ROTL
  Nach Links rotieren.
RUNF
  Vorwrts rennen. Entspricht FORW, jedoch sind die Zeiten etwas angepasst.

Die Zustnde der SMs sind jeweils in Unterschritte aufgeteilt und werden periodisch
durch den Overflow von Timer1 aktualisiert.
Timer1 und Timer2 erzeugen dabei die PWM-Signale fr die Servos.
Die seitlichen Servos sind an Timer1 angeschlossen und das mittlere an Timer2.
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

/*Konstanten fr die Servo-Positionen*/
#define LINKS_VORN 39
#define LINKS_MITTE 46
#define LINKS_HINTEN 53
#define HEBEN_LINKS 52
#define HEBEN_MITTE 44
#define HEBEN_RECHTS 36
#define RECHTS_VORN 52
#define RECHTS_MITTE 45
#define RECHTS_HINTEN 38

/*Haupt-FSM*/
unsigned char stateMain;
unsigned char stateMainStep;
unsigned short stateMainCnt;
#define STMAIN_NORMAL 0
#define STMAIN_AVOIDRIGHT 1
#define STMAIN_AVOIDLEFT 2
#define STMAIN_PANIC 3

/*Bewegungs-FSM*/
unsigned char stateWalk;
unsigned char stateWalkStep;
unsigned short stateWalkCnt;
unsigned char stateWalkNext;
#define STWALK_STOP 0
#define STWALK_FORW 1
#define STWALK_BACKW 2
#define STWALK_ROTR 3
#define STWALK_ROTL 4
#define STWALK_RUNF 5

/*Interval zwischen zwei Updates der SM in Millisekunden*/
#define STATEUPDATEINTERVAL 8

/*Initialisierungs-Routinen*/
void initPorts(void);
void initTimer(void);
void initSMs(void);

/*Routinen zur nderung der PWM-Signale*/
void links_vorn(void);
void links_mitte(void);
void links_hinten(void);
void heben_links(void);
void heben_mitte(void);
void heben_rechts(void);
void rechts_vorn(void);
void rechts_mitte(void);
void rechts_hinten(void);

int main(void)
{
  initPorts();
  initTimer();
  initSMs();

  sei();

  DDRC |= (1<<2);

  for(;;)
  {
    if(PINC & (1<<0))
    {
      PORTC |= (1<<2);
    }
    else
    {
      PORTC &= ~(1<<2);
      stateWalkNext = STWALK_ROTR;
    }
  }
}

void initPorts(void)
{
  /*Ausgnge fr Servos*/
  DDRB |= (1<<1)|(1<<2)|(1<<3);

  /*Pull-Ups anschalten*/
  PORTB = 0xff;
  PORTC = 0xff;
  PORTD = 0xff;
}

void initTimer(void)
{
  OCR1A = LINKS_MITTE;
  OCR1B = RECHTS_MITTE;
  TCCR1A = (1<<COM1A1)|(1<<COM1B1)|(1<<WGM10);
  TCCR1B = (1<<CS12);
  OCR2 = HEBEN_MITTE;
  TCCR2 = (1<<WGM20)|(1<<COM21)|(1<<CS22)|(1<<CS21);

  TIMSK |= (1<<TOIE1);
}

void initSMs(void)
{
  stateMain = STMAIN_NORMAL;

  stateWalk = STWALK_FORW;
  stateWalkStep = 3;
  stateWalkCnt = 0;
  stateWalkNext = STWALK_FORW;
}

void links_vorn(void)
{
  OCR1A = LINKS_VORN;
}

void links_mitte(void)
{
  OCR1A = LINKS_MITTE;
}

void links_hinten(void)
{
  OCR1A = LINKS_HINTEN;
}

void heben_links(void)
{
  OCR2 = HEBEN_LINKS;
}

void heben_mitte(void)
{
  OCR2 = HEBEN_MITTE;
}

void heben_rechts(void)
{
  OCR2 = HEBEN_RECHTS;
}

void rechts_vorn(void)
{
  OCR1B = RECHTS_VORN;
}

void rechts_mitte(void)
{
  OCR1B = RECHTS_MITTE;
}

void rechts_hinten(void)
{
  OCR1B = RECHTS_HINTEN;
}

/*ISR fr die Aktualisierung der SMs*/
SIGNAL(SIG_OVERFLOW1)
{
  /*aktueller Schritt vollstndig?*/
  if(stateWalkCnt > 0)
  {
    stateWalkCnt --;
  }
  else
  {
    /*auf bergang zu anderen Zustand testen*/
    if(stateWalk == STWALK_FORW && stateWalkNext == STWALK_ROTR)
    {
      if(stateWalkStep == 1)
      {
        stateWalk = STWALK_ROTR;
        stateWalkStep = 0;
      }
      else if(stateWalkStep == 2)
      {
        stateWalk = STWALK_ROTR;
        stateWalkStep = 2;
      }
    }
    else if(stateWalk == STWALK_ROTR && stateWalkNext == STWALK_FORW)
    {
      if(stateWalkStep == 1)
      {
        stateWalk = STWALK_FORW;
        stateWalkStep = 0;
      }
      else if(stateWalkStep == 3)
      {
        stateWalk = STWALK_FORW;
        stateWalkStep = 1;
      }
    }

    /*bei vorwrts und rechts rotieren: Schritt erhhen*/
    if(stateWalk == STWALK_FORW || stateWalk == STWALK_RUNF || stateWalk == STWALK_ROTR)
    {
      if(stateWalkStep < 3)
      {
        stateWalkStep ++;
      }
      else
      {
        stateWalkStep = 0;
      }
    }
    /*bei rckwrts und links rotieren: Schritt verringern*/
    if(stateWalk == STWALK_BACKW || stateWalk == STWALK_ROTL)
    {
      if(stateWalkStep > 0)
      {
        stateWalkStep --;
      }
      else
      {
        stateWalkStep = 3;
      }
    }

    /*vorwrts, rckwrts, runf nutzen die selbe SM*/
    if(stateWalk == STWALK_FORW || stateWalk == STWALK_RUNF || stateWalk == STWALK_BACKW)
    {
      if(stateWalkStep == 0)
      {
        heben_links();
        if(stateWalk == STWALK_RUNF)
        {
          stateWalkCnt = 80 / STATEUPDATEINTERVAL;
        }
        else
        {
          stateWalkCnt = 180 / STATEUPDATEINTERVAL;
        }
      }
      else if(stateWalkStep == 1)
      {
        links_vorn();
        rechts_hinten();
        stateWalkCnt = 220 / STATEUPDATEINTERVAL;
      }
      else if(stateWalkStep == 2)
      {
        heben_rechts();
        if(stateWalk == STWALK_RUNF)
        {
          stateWalkCnt = 80 / STATEUPDATEINTERVAL;
        }
        else
        {
          stateWalkCnt = 180 / STATEUPDATEINTERVAL;
        }
      }
      else if(stateWalkStep == 3)
      {
        links_hinten();
        rechts_vorn();
        stateWalkCnt = 220 / STATEUPDATEINTERVAL;
      }
    }

    /*rechts und links rotieren nutzen die selbe SM*/
    if(stateWalk == STWALK_ROTR || stateWalk == STWALK_ROTL)
    {
      if(stateWalkStep == 0)
      {
        heben_links();
        stateWalkCnt = 180 / STATEUPDATEINTERVAL;
      }
      else if(stateWalkStep == 1)
      {
        links_vorn();
        rechts_vorn();
        stateWalkCnt = 220 / STATEUPDATEINTERVAL;
      }
      else if(stateWalkStep == 2)
      {
        heben_rechts();
        stateWalkCnt = 180 / STATEUPDATEINTERVAL;
      }
      else if(stateWalkStep == 3)
      {
        links_hinten();
        rechts_hinten();
        stateWalkCnt = 220 / STATEUPDATEINTERVAL;
     }
    }
  }
}
